"""Utilities for configuring `aga`."""

from dataclasses import MISSING, dataclass, field, fields
from importlib.resources import files
from typing import Any

import toml
from dacite import from_dict  # type: ignore


def _default_value(path: list[str]) -> Any:
    """Given a path of config names, get the value of the default at the leaf."""
    curr = _DEFAULT_CONFIG
    for val in path:
        curr = getattr(curr, val)

    return curr


def _from_default(path: list[str]) -> Any:
    """Given a path of config names, construct a field which inherits the default."""
    return field(default_factory=lambda: _default_value(path))  # type: ignore


@dataclass
class AgaTestConfig:
    """Aga's test-related configuration."""

    name_sep: str = _from_default(["test", "name_sep"])
    name_fmt: str = _from_default(["test", "name_fmt"])
    failure_msg: str = _from_default(["test", "failure_msg"])
    error_msg: str = _from_default(["test", "error_msg"])
    stdout_differ_msg: str = _from_default(["test", "stdout_differ_msg"])
    diff_explanation_msg: str = _from_default(["test", "diff_explanation_msg"])

    def update_weak(self, other: "AgaTestConfig") -> None:
        """Update all default attributes of self to match other."""
        _update_weak_leaf(self, other)


@dataclass
class AgaLoaderConfig:
    """Aga's submission loading configuration."""

    # this is pretty gross, we should find an easier way to do this
    # we're doing this so we can single-source-of-truth the defaults in `defults.toml`.
    import_error_msg: str = _from_default(["loader", "import_error_msg"])
    no_match_msg: str = _from_default(["loader", "no_match_msg"])
    too_many_matches_msg: str = _from_default(["loader", "too_many_matches_msg"])
    no_script_error_msg: str = _from_default(["loader", "no_script_error_msg"])
    multiple_scripts_error_msg: str = _from_default(
        ["loader", "multiple_scripts_error_msg"]
    )

    def update_weak(self, other: "AgaSubmissionConfig") -> None:
        """Update all default attributes of self to match other."""
        _update_weak_leaf(self, other)


@dataclass
class AgaSubmissionConfig:
    """Aga's submission configuration."""

    failed_tests_msg: str = _from_default(["submission", "failed_tests_msg"])
    failed_hidden_tests_msg: str = _from_default(
        ["submission", "failed_hidden_tests_msg"]
    )
    no_failed_tests_msg: str = _from_default(["submission", "no_failed_tests_msg"])

    def update_weak(self, other: "AgaSubmissionConfig") -> None:
        """Update all default attributes of self to match other."""
        _update_weak_leaf(self, other)


@dataclass
class AgaProblemConfig:
    """Aga's problem-related configuration."""

    check_stdout: bool = _from_default(["problem", "check_stdout"])
    check_stdout_overridden: bool = False

    mock_input: bool = _from_default(["problem", "mock_input"])
    mock_input_overridden: bool = False

    def update_weak(self, other: "AgaProblemConfig") -> None:
        """Update all default attributes of self to match other."""
        _update_weak_leaf(self, other)


@dataclass
class AgaConfig:
    """Aga's configuration."""

    test: AgaTestConfig = field(default_factory=AgaTestConfig)
    submission: AgaSubmissionConfig = field(default_factory=AgaSubmissionConfig)
    loader: AgaLoaderConfig = field(default_factory=AgaLoaderConfig)
    problem: AgaProblemConfig = field(default_factory=AgaProblemConfig)

    def update_weak(self, other: "AgaConfig") -> None:
        """Update all default attributes of self to match other."""
        for attr in fields(self):
            getattr(self, attr.name).update_weak(getattr(other, attr.name))


def _update_weak_leaf(self, other) -> None:  # type: ignore
    """Update a leaf-level dataclass weakly."""
    # iterate over all the fields of the dataclass, checking them against the
    # default value, and updating them to the other value if they match the default
    # value
    for attr in fields(self):
        if (
            attr.default_factory != MISSING  # type: ignore
            # flake8: noqa
            and attr.default_factory() == getattr(self, attr.name)  # type: ignore
        ):
            try:
                if getattr(self, attr.name + "_overridden"):
                    continue
            except AttributeError:
                # if there is no overridden tracker, we assume it's safe to override
                pass

            setattr(self, attr.name, getattr(other, attr.name))


def load_config_from_path(path: str) -> AgaConfig:
    """Load the toml file at path."""
    return from_dict(  # type: ignore
        data_class=AgaConfig,
        data=toml.load(path),
    )


_DEFAULT_CONFIG: AgaConfig = load_config_from_path(
    # this is type safe, but there's not actually a safe way to indicate that to mypy,
    # since `Traversable` (the type returned by `joinpath`) isn't re-exported by
    # `importlib.resources`
    files("aga").joinpath("defaults.toml")  # type: ignore
)



"""Tools for loading student-submitted, maybe-invalid code from source files."""

import importlib.util
import os
from glob import glob
from importlib.machinery import ModuleSpec
from os.path import isdir
from os.path import join as pathjoin
from types import ModuleType
from typing import Any, Callable, Iterable, TypeVar

from dill import Unpickler  # type: ignore

from .core import Problem

Output = TypeVar("Output")


class InvalidSubmissionError(BaseException):
    """Something about the submission was invalid."""


class TooManyMatchingSymbols(InvalidSubmissionError):
    """Too many maching symbols were found."""


class NoMatchingSymbol(InvalidSubmissionError, AttributeError):
    """An expected symbol was not found."""


class SubmissionSyntaxError(InvalidSubmissionError, SyntaxError):
    """The submission held an invalid syntax."""

    def __init__(self, file: str):
        super().__init__()
        self.file = file


class NoScript(InvalidSubmissionError, FileNotFoundError):
    """No script was found."""


class MultipleScripts(InvalidSubmissionError):
    """Too many scripts were found."""


def _get_spec_from_path(path: str, name: str) -> ModuleSpec:
    """Get the spec of the module at path."""
    spec = importlib.util.spec_from_file_location(name, path)

    if spec is None:
        # Based on inspection of the source, I'm not certain how this can happen, but my
        # type checker insists it can. This seems like the most reasonable error to
        # raise.
        raise FileNotFoundError

    return spec


def _load_script_from_file(path: str, name: str = "script") -> Callable[[], None]:
    """Load the python script at path as a function."""
    spec = _get_spec_from_path(path, name)
    mod = importlib.util.module_from_spec(spec)

    def inner() -> None:
        # we don't handle errors here so that they'll show up as runtime issues when we
        # run the test
        spec.loader.exec_module(mod)  # type: ignore

    return inner


def _load_script_from_dir(path: str, name: str = "script") -> Callable[[], None]:
    """Load the python script in the directory at path."""
    scripts = glob(pathjoin(path, "*.py"))
    if not scripts:
        raise NoScript
    if len(scripts) > 1:
        raise MultipleScripts

    return _load_script_from_file(scripts[0], name)


def load_script_from_path(path: str, name: str = "script") -> Callable[[], None]:
    """Load the script at path, either a file or directory."""
    if isdir(path):
        return _load_script_from_dir(path, name)
    else:
        return _load_script_from_file(path, name)


def _load_source_from_file(path: str, name: str = "module") -> Any:
    """Load the python source file found at path, absolute or relative, as a module.

    There's a lot of weird stuff going on in this method with type signatures and
    poorly-documented code that python uses internally for their `import` statement. I
    got this implementation from https://stackoverflow.com/a/67692 and made only small
    modifications to it, but I'm not 100% sure I can explain how it works.
    """
    spec = _get_spec_from_path(path, name)
    mod = importlib.util.module_from_spec(spec)

    try:
        spec.loader.exec_module(mod)  # type: ignore
    except (SyntaxError, NameError) as err:
        raise SubmissionSyntaxError(path) from err  # group all parse errors

    return mod


def _load_attr_from_module(attr: str, module: ModuleType) -> Any:
    """Get a specific symbol from a module."""
    try:
        return module.__getattribute__(attr)
    except AttributeError as err:
        raise NoMatchingSymbol from err


def _load_from_module_by(
    pred: Callable[[Any], bool], module: ModuleType
) -> Iterable[Any]:
    """Return all items in the module satisfying the predicate."""
    for item in module.__dict__.values():
        if pred(item):
            yield item


def _load_problems_from_module(module: ModuleType) -> Iterable[Problem[Any]]:
    """Return all problems in the module."""
    yield from _load_from_module_by(
        lambda i: isinstance(i, Problem), module  # type: ignore
    )


def load_problems_from_path(path: str) -> Iterable[Problem[Any]]:
    """Load all problems from the module at path."""
    mod = _load_source_from_file(path)
    yield from _load_problems_from_module(mod)


def _load_symbol_from_file(path: str, symbol: str) -> Any:
    """Load a specific symbol from a source file found at path, absolute or relative."""
    mod = _load_source_from_file(path)
    return _load_attr_from_module(symbol, mod)


def _load_symbol_from_dir(path: str, symbol: str) -> Any:
    """Load a specific symbol from any of the source files in a directory."""
    matching_symbols = []
    for file in os.listdir(path):

        try:
            file_path = pathjoin(path, file)
            matching_symbols.append(load_symbol_from_path(file_path, symbol))
        except (FileNotFoundError, NoMatchingSymbol):
            pass

    if len(matching_symbols) > 1:
        raise TooManyMatchingSymbols
    if len(matching_symbols) == 0:
        raise NoMatchingSymbol
    return matching_symbols[0]


def load_symbol_from_path(path: str, symbol: str) -> Any:
    """Load a specific symbol from `path`.

    If path is a directory, load from any file in the directory. If path is a file, load
    from that file.
    """
    if isdir(path):
        return _load_symbol_from_dir(path, symbol)
    else:
        return _load_symbol_from_file(path, symbol)


class _ProblemUnpickler(Unpickler):  # type: ignore
    """A custom unpickler which will always get the `Problem` class from `aga`.

    This is a hack-ish thing which is required because dill expects us to unpickle an
    object in the same module it was pickled in, so it can then find the object's type
    and use that for instantiation. We want to be able to unpickle the object in any
    type, and we know that we always have a Problem pickled at `problem.pckl`, so we can
    just assert that its class should be Problem. This is _highly_ unsafe if we are
    unsure of the safety of `problem.pckl`, but pickle/dill is not remotely safe anyway
    with untrusted data.

    This specific solution will break if dill, for some reason, wants to pickle some
    *other* class named "Problem". In that case, I think the best solution will be to
    look into a custom pickler which changes the module name on that end.
    """

    def find_class(self, module: str, name: str) -> Any:
        if name == "Problem":
            return Problem
        return super().find_class(module, name)


def load_problem(root: str, fname: str = "problem.pckl") -> Problem[Output]:
    """Load a problem from the gradescope environment."""
    with open(pathjoin(root, fname), "rb") as problem_pickled:
        out: Problem[Output] = _ProblemUnpickler(problem_pickled).load()
    return out



"""Additional checks and filters for problems."""

from io import StringIO
from typing import Any, Callable, TypeVar
from unittest.mock import patch

Output = TypeVar("Output")


def with_captured_stdout(
    func: Callable[..., Output]
) -> Callable[..., tuple[str, Output]]:
    """Run func, returning its stdout and normal return value."""

    def inner(*args: Any, **kwargs: Any) -> tuple[str, Any]:
        with patch("sys.stdout", new_callable=StringIO) as stdout:
            func_out = func(*args, **kwargs)

        return stdout.getvalue(), func_out

    return inner



"""Aga grades assignments: a library for easily producing autograders for code.

Anything not explicitly documented here should not be used directly by clients and is
only exposed for testing, the CLI, and type hinting.
"""

from importlib.metadata import version

from .core import group, problem, test_case, test_cases

__all__ = ("problem", "test_case", "test_cases", "group")
__version__ = version(__name__)  # type: ignore



"""The core entrypoint for the gradescope frontend.

More complete documentation is in `__init__.py`.
"""

from importlib.abc import Traversable
from importlib.resources import files
from os import makedirs
from os.path import join as pathjoin
from shutil import copyfileobj
from tempfile import TemporaryDirectory
from typing import Iterable, Optional
from zipfile import ZipFile

from dill import dump  # type: ignore

from ..core import Output, Problem

# don't zip resources because we handle them manually
_ZIP_IGNORES = ("__pycache__", "resources")

# files we need present in the gradescope environment that aren't part of the `aga`
# source or of the specific problem
_GS_ENV_DEPS = ("setup.sh", "setup.py", "run_autograder")


class InvalidProblem(BaseException):
    """The Problem failed some golden tests."""


def into_gradescope_zip(problem: Problem[Output], path: Optional[str] = None) -> str:
    """Convert a Problem into a gradescope autograder zip, returning its location.

    This is the high-level entrypoint for this module.
    """
    _check_problem(problem)
    zip_name = _get_zipfile_dest(path, problem)

    with TemporaryDirectory() as tempdir:
        with ZipFile(zip_name, "w") as zip_f:
            for (src_path, resource_name) in _copy_package_to(tempdir, files("aga")):
                zip_f.write(src_path, arcname=resource_name)

            for file in _GS_ENV_DEPS:
                path = _manual_copy_resource_to(tempdir, file)
                zip_f.write(path, arcname=file)

            path = _dump_problem_into_dir(problem, tempdir)
            zip_f.write(path, arcname="problem.pckl")

    return zip_name


def _get_zipfile_dest(path: Optional[str], problem: Problem[Output]) -> str:
    """Determine the destination in which to put the zip file.

    If `path` is none, this is the problem's name; otherwise it's just the provided
    path.
    """
    if path is None:
        return problem.name() + ".zip"
    else:
        return path


def _check_problem(problem: Problem[Output]) -> None:
    """Check whether `problem` is valid.

    Currently, this just runs the golden tests for problem.
    """
    try:
        problem.check()
    except AssertionError as err:
        raise InvalidProblem from err


def _copy_package_to(
    tempdir: str, package: Traversable, prefix: str = "aga"
) -> Iterable[tuple[str, str]]:
    """Recusrively copy each resource, yielding the file path and resource name."""
    # can we use `copytree` from shutil for this? or maybe `inspect.getsource`?

    makedirs(pathjoin(tempdir, prefix), exist_ok=True)
    # manually traverse the package structure, copying source files as we go
    for resource in package.iterdir():
        if resource.name not in _ZIP_IGNORES:
            if resource.is_dir():
                yield from _copy_package_to(
                    tempdir,
                    resource,
                    pathjoin(prefix, resource.name),
                )
            else:
                yield _copy_resource_to(tempdir, resource, prefix)


def _copy_resource_to(
    tempdir: str, resource: Traversable, prefix: str
) -> tuple[str, str]:
    """Copy the resource to tempdir, yielding the file path and resource name."""
    dest_path = pathjoin(tempdir, prefix, resource.name)
    with resource.open() as src:  # type: ignore
        with open(dest_path, "w", encoding="UTF-8") as dest:
            copyfileobj(src, dest)

    return (dest_path, pathjoin(prefix, resource.name))


def _manual_copy_resource_to(
    tempdir: str, fname: str, package: str = "aga.gradescope.resources"
) -> str:
    """Copy the resource at package.fname to tempdir/fname, returning the dest path."""
    dest_path = pathjoin(tempdir, fname)
    with files(package).joinpath(fname).open() as src:  # type: ignore
        with open(dest_path, "w", encoding="UTF-8") as dest:
            copyfileobj(src, dest)

    return dest_path


def _dump_problem_into_dir(
    problem: Problem[Output], tempdir: str, fname: str = "problem.pckl"
) -> str:
    """Dump a problem into a directory, returning the pckl file path."""
    path = pathjoin(tempdir, fname)
    _dump_problem_at_path(problem, path)
    return path


def _dump_problem_at_path(problem: Problem[Output], dest: str) -> None:
    """Pickle the problem into a destination."""
    with open(dest, "wb") as file:
        dump(problem, file)



"""Install the library for usage on gradescope.

This is _not_ a permanent solution, it's just a hack so we don't need to continually
push to PyPI while developing. (Even Test PyPI doesn't really work, it's not designed
for this sort of thing and you don't want to have to push to test local changes.) We
should either:

1. Find a way to automatically generate `setup.py`, maybe poetry has an API for this but
I couldn't find anything.
2. Find a better way to install the package from in gradescope without needing to fetch
it from PyPI.
"""

from setuptools import setup  # type: ignore

setup(
    name="aga",
    version="0.8.6",
    install_requires=[
        "dataclasses-json >= 0.5.6",
        "dill >= 0.3.4",
        "gradescope-utils >= 0.4.0",
        "typer >= 0.4.0",
        "dacite >= ^1.6.0",
        "toml >= ^0.10.2",
        "types-toml >= ^0.10.8",
    ],
)



"""Methods for running tests.

The main work of this module is to turn a test suite, generated by the core module, into
an `ProblemOutput`, which contains data related to test success or failure. It
provides the `run` method, which does so.

For convenience, it also provides the `load_and_run` method, which loads a student
submission and then runs it.
"""

from dataclasses import dataclass
from typing import Any, Optional
from unittest import TestResult

from .config import AgaConfig
from .core import AgaTestCase, AgaTestSuite, Output, Problem, SubmissionMetadata
from .loader import (
    MultipleScripts,
    NoMatchingSymbol,
    NoScript,
    SubmissionSyntaxError,
    TooManyMatchingSymbols,
    load_script_from_path,
    load_symbol_from_path,
)
from .score import ScoredPrize
from .util import limited_traceback


@dataclass
class TcOutput:
    """Stores information about a completed test case.

    Attributes
    ----------
    score : float
        The test's score.
    max_score : float
        The max score for the test.
    name : str
        The test's name.
    output : Optional[str]
        Human-readable text output of the test. Some frontends distinguish between no
        output and empty output, i.e. in terms of showing UI elements.
    hidden : bool
        The test's visibility.
    """

    score: float
    max_score: float
    name: str
    output: Optional[str] = None
    hidden: bool = False

    def is_correct(self) -> bool:
        """Check whether the problem recieved full credit."""
        return self.score == self.max_score


@dataclass
class ProblemOutput:
    """Stores information about a completed problem.

    Attributes
    ----------
    score : float
        The total score for the problem. We currently do not track max score,
        but instead expect frontends to do so.
    tests : list[TcOutput]
        The test cases for the problem.
    output : str
        Human-readable text output of the problem.
    """

    tests: list[TcOutput]
    score: float
    output: str


class _AgaTestResult(TestResult):
    """A custom `TestResult` that constructs the gradescope json object for the test.

    This is _not_ safe to be used in place of an arbitrary TestResult. This is because
    it relies on specific methods of the `AgaTestCase` class. It should only be used
    with TestSuites that we _know_ only containt `AgaTestCase`s.
    """

    def __init__(
        self,
        config: AgaConfig,
        prizes: list[ScoredPrize],
        metadata: SubmissionMetadata,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        super().__init__(*args, **kwargs)
        self._tests: list[TcOutput] = []
        self._output_msgs: list[str] = []
        self._config = config
        self._prizes = prizes
        self._metadata = metadata

    @staticmethod
    def _test_data(test: AgaTestCase) -> TcOutput:
        """Construct the test data for a successful test, with _no_ output."""
        metadata = test.metadata()

        return TcOutput(
            max_score=metadata.max_score,
            name=metadata.name,
            score=metadata.max_score,
            hidden=metadata.hidden,
        )

    def _fail_data(self, test: AgaTestCase, err) -> TcOutput:  # type: ignore
        """Construct the test data for a failure."""
        data = self._test_data(test)
        data.output = str(err[1])
        data.score = 0.0

        return data

    def _err_data(self, test: AgaTestCase, err) -> TcOutput:  # type: ignore
        """Construct the test data for an error."""
        data = self._test_data(test)
        data.output = test.metadata().config.error_msg.format(
            type=err[0].__name__,
            message=err[1],
            traceback=limited_traceback(err[2]),
        )
        data.score = 0.0

        return data

    # lots of type: ignores because these methods all restrict the input TestCase type

    def addError(self, test: AgaTestCase, err) -> None:  # type: ignore
        """Add an error."""
        super().addError(test, err)
        self._tests.append(self._err_data(test, err))

    def addFailure(self, test: AgaTestCase, err) -> None:  # type: ignore
        """Add a failure."""
        super().addFailure(test, err)
        self._tests.append(self._fail_data(test, err))

    def addSuccess(self, test: AgaTestCase) -> None:  # type: ignore[override]
        """Add a success."""
        super().addSuccess(test)
        self._tests.append(self._test_data(test))

    def _score(self) -> float:
        """Get the total score."""
        return sum(t.score for t in self._tests)

    def _output(self) -> str:
        """Get the current output string.

        Generally, `build_output` should be called once, and then this method used for
        subsequent gets.
        """
        return "\n\n".join(self._output_msgs)

    def _add_prizes(self) -> None:
        prize_tests = []

        for prize in self._prizes:
            # mypy bug (https://github.com/python/mypy/issues/5485, fixed on main)
            scalar, message = prize.prize.criteria(  # type: ignore
                self._tests, self._metadata
            )
            score = scalar * prize.max_score
            prize_out = TcOutput(
                score=score,
                max_score=prize.max_score,
                name=prize.prize.name,
                output=message,
            )

            # don't append to self._tests immediately so the next prizes don't see this
            # one
            prize_tests.append(prize_out)

        self._tests += prize_tests

    def _build_output(self) -> str:
        """Build the main output string."""
        config = self._config.submission

        failed_tests = [t for t in self._tests if not t.is_correct()]
        if failed_tests:
            # add failed test message
            self._output_msgs.append(config.failed_tests_msg)

            if any(t.hidden for t in failed_tests):
                # add hidden test message
                self._output_msgs.append(config.failed_hidden_tests_msg)

        else:
            self._output_msgs.append(config.no_failed_tests_msg)

        return self._output()

    def build(self) -> ProblemOutput:
        """Get the completed output data."""
        output = self._build_output()
        # we run this after so the tests failed message gets built without taking prizes
        # into account
        self._add_prizes()

        return ProblemOutput(tests=self._tests, score=self._score(), output=output)


def _run(
    suite: AgaTestSuite, prizes: list[ScoredPrize], metadata: SubmissionMetadata
) -> ProblemOutput:
    """Run the suite, returning the output."""
    return suite.run(  # type: ignore
        _AgaTestResult(suite.config, prizes, metadata)
    ).build()


def load_and_run(
    problem: Problem[Output],
    path: str,
    metadata: SubmissionMetadata,
) -> ProblemOutput:
    """Load the submission and then run the suite, returning the output.

    The path can be either a directory, which will be searched without recurring into
    subdirctories, or a single file. This method handles errors from missing or invalid
    submissions.
    """
    try:
        if not problem.is_script:
            under_test = load_symbol_from_path(path, problem.expected_symbol())
        else:
            under_test = load_script_from_path(path)
    except SubmissionSyntaxError as err:
        return ProblemOutput(
            output=_submission_syntax_error_msg(
                err.__cause__, problem.config()  # type: ignore
            ),
            tests=[],
            score=0.0,
        )
    except NoMatchingSymbol:
        return ProblemOutput(
            output=_no_matches_error_msg(problem),
            tests=[],
            score=0.0,
        )
    except TooManyMatchingSymbols:
        return ProblemOutput(
            output=_too_many_matches_error_msg(problem),
            tests=[],
            score=0.0,
        )
    except NoScript:
        return ProblemOutput(
            output=_no_script_error_msg(problem),
            tests=[],
            score=0.0,
        )
    except MultipleScripts:
        return ProblemOutput(
            output=_multiple_scripts_error_msg(problem),
            tests=[],
            score=0.0,
        )

    suite, prizes = problem.generate_test_suite(under_test, metadata)
    return _run(suite, prizes, metadata)


def _submission_syntax_error_msg(cause: SyntaxError, config: AgaConfig) -> str:
    return config.loader.import_error_msg.format(message=str(cause))


def _no_matches_error_msg(problem: Problem[Output]) -> str:
    return problem.config().loader.no_match_msg.format(name=problem.expected_symbol())


def _too_many_matches_error_msg(problem: Problem[Output]) -> str:
    return problem.config().loader.too_many_matches_msg.format(
        name=problem.expected_symbol()
    )


def _no_script_error_msg(problem: Problem[Output]) -> str:
    return problem.config().loader.no_script_error_msg


def _multiple_scripts_error_msg(problem: Problem[Output]) -> str:
    return problem.config().loader.multiple_scripts_error_msg



# pylint: disable=invalid-name,too-many-instance-attributes
# These are mandated by the json schema
"""Parses Gradescope's `submission_metadata.json` file."""

from dataclasses import dataclass, field
from datetime import datetime
from typing import List, Optional

from dataclasses_json import config, dataclass_json
from marshmallow import fields

from ..core import SubmissionMetadata
from .schema import GradescopeJson


@dataclass_json
@dataclass(frozen=True)
class GradescopeAssignmentMetadata:
    """The JSON schema for Gradescope's assignment settings.

    Attributes
    ----------
    due_date : datetime
        The assignment's due date.
    group_size : Optional[int]
        The maximum group size on a group assignment.
    group_submission : bool
        Whether group submission is allowed.
    id : int
        The assignment's ID.
    course_id : int
        The course's ID.
    late_due_date : Optional[datetime]
        The late due date, or None if late submission disallowed.
    release_date : datetime
        The assignment's release date.
    title : str
        The assignment's title.
    total_points : float
        The total point value of the assignment.
    """

    due_date: datetime = field(
        metadata=config(
            encoder=datetime.isoformat,
            decoder=datetime.fromisoformat,  # type: ignore
            mm_field=fields.DateTime(format="iso"),
        )
    )
    group_size: Optional[int]
    group_submission: bool
    id: int
    course_id: int
    late_due_date: Optional[datetime] = field(
        metadata=config(
            encoder=lambda s: datetime.isoformat(s) if s else None,
            decoder=lambda s: datetime.fromisoformat(s) if s else None,  # type: ignore
            mm_field=fields.DateTime(format="iso"),
        )
    )
    release_date: datetime = field(
        metadata=config(
            encoder=datetime.isoformat,
            decoder=datetime.fromisoformat,  # type: ignore
            mm_field=fields.DateTime(format="iso"),
        )
    )
    title: str
    total_points: float = field(metadata=config(encoder=str, decoder=float))


@dataclass_json
@dataclass(frozen=True)
class GradescopeAssignmentUser:
    """The JSON schema for a 'user' (submitter) of a Gradescope assignment.

    Attributes
    ----------
    email : str
        The submitter's email.
    id : int
        The submitter's ID.
    name : str
        The submitter's name.
    """

    email: str
    id: int
    name: str


@dataclass_json
@dataclass(frozen=True)
class GradescopePreviousSubmission:
    """The JSON schema for a previous submission record.

    Attributes
    ----------
    submission_time : datetime
        The time of the previous submission.
    score : float
        The previous submission's score.
    results : GradescopeJson
        The results.json file from the previous submission.
    """

    submission_time: datetime = field(
        metadata=config(
            encoder=datetime.isoformat,
            decoder=datetime.fromisoformat,  # type: ignore
            mm_field=fields.DateTime(format="iso"),
        )
    )
    score: float
    results: GradescopeJson


@dataclass_json
@dataclass(frozen=True)
class GradescopeSubmissionMetadata:
    """The JSON schema for Gradescope's submission metadata.

    Documentation online here
    <https://gradescope-autograders.readthedocs.io/en/latest/submission_metadata/>.

    Attributes
    ----------
    id : int
        A unique submission identifier.
    created_at : datetime
        The submission time.
    assignment : GradescopeAssignmentMetadata
        The assignment's specific settings.
    submission_method : str
        One of "upload", "GitHub", or "Bitbucket"
    users : List[GradescopeAssignmentUser]
        The people who submitted the assignment. Multiple if the assignment is a group
        assignment.
    previous_submissions : List[GradescopePreviousSubmission]
        The previous submissions.
    """

    id: int
    created_at: datetime = field(
        metadata=config(
            encoder=datetime.isoformat,
            decoder=datetime.fromisoformat,  # type: ignore
            mm_field=fields.DateTime(format="iso"),
        )
    )
    assignment: GradescopeAssignmentMetadata
    submission_method: str
    users: List[GradescopeAssignmentUser]
    previous_submissions: List[GradescopePreviousSubmission]

    def to_submission_metadata(self) -> SubmissionMetadata:
        """Get the aga submission metadata."""
        return SubmissionMetadata(
            total_score=self.assignment.total_points,
            time_since_due=self.created_at - self.assignment.due_date,
            previous_submissions=len(self.previous_submissions),
        )


def load_submission_metadata_from_path(path: str) -> GradescopeSubmissionMetadata:
    """Load the submission metadata from a json file."""
    with open(path, encoding="UTF-8") as file:
        # pylint: disable=no-member
        return GradescopeSubmissionMetadata.from_json(file.read())  # type: ignore



"""Contains a TestResult and TestRunner for producing gradescope's results.json file.

The design of this module is inspire by gradescope_utils's JSONTestRunner, but this is a
complete rewrite.
"""

from dataclasses import dataclass, field
from typing import Optional, TextIO

from dataclasses_json import dataclass_json

from ..runner import ProblemOutput, TcOutput


@dataclass_json
@dataclass
class GradescopeTestJson:
    """The JSON schema for a single Test.

    Attributes
    ----------
    score : Optional[float]
        The test's score. Required if no top-level score is set.
    max_score : Optional[float]
        The max score for the test.
    name : Optional[str]
        The test's name.
    output : Optional[str]
        Human-readable text output of the test.
    tags : Optional[List[str]]
        Tags for the test.
    visibility : str
        The test's visibility. "hidden", "visible", "after_due_date", "after_published"
    """

    score: Optional[float] = None
    max_score: Optional[float] = None
    name: Optional[str] = None
    number: Optional[float] = None
    output: Optional[str] = None
    tags: Optional[str] = None
    visibility: Optional[str] = None

    @classmethod
    def from_test_case_output(cls, output: TcOutput) -> "GradescopeTestJson":
        """Create a test json from a test case output."""
        return cls(
            score=output.score,
            max_score=output.max_score,
            name=output.name,
            output=output.output,
            visibility="hidden" if output.hidden else "visible",
        )


@dataclass_json
@dataclass
class GradescopeJson:
    """The JSON schema for Gradescope.

    We currently don't support the leaderboard and extra_data features of the gradescope
    schema. Those are documented on the autograder documentation, here:
    <https://gradescope-autograders.readthedocs.io/en/latest/specs/>.

    Attributes
    ----------
    tests : List[_GradescopeJsonTest]
        The tests for the problem. Required if no global score provided.
    score : Optional[float]
        The overall score. Required if any test has no set score.
    execution_time : Optional[int]
        The execution time of all the tests, in seconds.
    output : Optional[str]
        The top-level, human-readable text output for all the problems.
    visibility : Optional[str]
        The default visibility for each test. Overridden by test-specific settings.
    stdout_visibility : Optional[str]
        Whether to show stdout for the tests. Same options as for visibility.
    """

    tests: list[GradescopeTestJson] = field(default_factory=list)
    score: Optional[float] = None
    execution_time: Optional[int] = None
    output: Optional[str] = None
    visibility: Optional[str] = None
    stdout_visibility: Optional[str] = None

    @classmethod
    def from_problem_output(cls, output: ProblemOutput) -> "GradescopeJson":
        """Create a gradescop json from a problem output."""
        return cls(
            tests=[GradescopeTestJson.from_test_case_output(tc) for tc in output.tests],
            score=output.score,
            output=output.output,
        )


def write_to(output: ProblemOutput, stream: TextIO) -> None:
    """Write the data from output as gradescope json to a text stream."""
    # pylint: disable=no-member
    json = GradescopeJson.from_problem_output(output).to_json()  # type: ignore
    stream.write(json)



"""Contains the `main` function which is used by the run_autograder script."""

from os.path import join as pathjoin
from typing import TypeVar

from ..loader import load_problem
from ..runner import load_and_run
from .metadata import load_submission_metadata_from_path
from .schema import write_to

Output = TypeVar("Output")

_AUTOGRADER_ROOT = "/autograder"
_AUTOGRADER_SRC = pathjoin(_AUTOGRADER_ROOT, "source")
_AUTOGRADER_SUBMISSION = pathjoin(_AUTOGRADER_ROOT, "submission")
_AUTOGRADER_METADATA = pathjoin(_AUTOGRADER_ROOT, "submission_metadata.json")
_OUTPUT_FILE = pathjoin(_AUTOGRADER_ROOT, "results/results.json")


def gradescope_main(
    problem_src: str = _AUTOGRADER_SRC,
    submission_dir: str = _AUTOGRADER_SUBMISSION,
    metadata_file: str = _AUTOGRADER_METADATA,
    output_file: str = _OUTPUT_FILE,
) -> None:
    """Load and run the tests against student code.

    The parameters are to allow overriding the file structure in unit testing.
    """
    problem = load_problem(problem_src)  # type: ignore
    metadata = load_submission_metadata_from_path(
        metadata_file
    ).to_submission_metadata()
    output = load_and_run(problem, submission_dir, metadata)

    with open(output_file, "w+", encoding="UTF-8") as file:
        write_to(output, file)



"""The core library functionality."""

from collections.abc import Iterable
from copy import deepcopy
from dataclasses import dataclass
from datetime import timedelta
from itertools import product
from typing import Any, Callable, Generic, Optional, TypeVar
from unittest import TestCase, TestSuite
from unittest.mock import patch

from .checks import with_captured_stdout
from .config import AgaConfig, AgaTestConfig
from .score import Prize, ScoredPrize, ScoreInfo, compute_scores
from .util import text_diff

Output = TypeVar("Output")


@dataclass(frozen=True)
class TestMetadata:
    """Stores metadata about a specific test case."""

    name: str
    max_score: float
    config: AgaTestConfig
    check_stdout: bool
    mock_input: bool
    hidden: bool = False


@dataclass(frozen=True)
class SubmissionMetadata:
    """Metadata for testing a submission, collected from the frontend.

    Attributes
    ----------
    total_score : float
        The problem's total score.
    time_since-due : timedelta
        The delta _from_ the due date _to_ the submission date, i.e. it's negative if
        the problem was submitted before the due date.
    previous_submissions : int
        The number of previous submissions.

    """

    total_score: float
    time_since_due: timedelta
    previous_submissions: int

    def is_on_time(self) -> bool:
        """Return true of the submission was on time."""
        return self.time_since_due <= timedelta()


class AgaTestCase(TestCase):
    """A `TestCase` which runs a single test of a `Problem`."""

    def __init__(
        self,
        test_input: "_TestInputs[Output]",
        golden: Callable[..., Output],
        under_test: Callable[..., Output],
        metadata: TestMetadata,
    ) -> None:
        super().__init__()
        self._test_input = test_input
        self._golden = golden
        self._under_test = under_test
        self._metadata = metadata

    def metadata(self) -> TestMetadata:
        """Get the test's metadata."""
        return self._metadata

    # pylint: disable=invalid-name
    # camelCase is required by unittest
    def runTest(self) -> None:
        """Run the test case."""
        self._test_input.check(self._golden, self._under_test, self._metadata)

    # pylint: disable=invalid-name
    # camelCase is required by unittest
    def shortDescription(self) -> str:
        """Dynamically generate the test name.

        This method is called by unittest.
        """
        return self._metadata.name


class AgaTestSuite(TestSuite):
    """A thin wrapper around TestSuite that store a config."""

    def __init__(self, config: AgaConfig, tests: list[AgaTestCase]):
        super().__init__(tests)
        self.config = config


class _TestInputs(TestCase, Generic[Output]):
    """A single set of test inputs for a problem.

    These will be run against the golden solution and the function under test; their
    outputs will be compared, and a unittest failure raised if they differ.
    """

    # this tells unittest not to print their default assertion error messages
    longMessage = False

    def __init__(
        self,
        *args: Any,
        aga_expect: Optional[Output],
        aga_hidden: bool,
        aga_name: Optional[str],
        aga_weight: int,
        aga_value: float,
        aga_mock_input: bool,
        **kwargs: Any,
    ) -> None:
        super().__init__()
        self._name = aga_name
        self._hidden = aga_hidden
        self._mock_input = aga_mock_input
        self._expect = aga_expect
        self.score_info = ScoreInfo(aga_weight, aga_value)

        self._args = args
        self._kwargs = kwargs

    def _eval(self, func: Callable[..., Any]) -> Any:
        """Evaluate func on the arguments."""
        # deepcopy in case the student submission mutates arguments; we don't want it to
        # mess with our copy of the arguments
        if self._mock_input:
            with patch("builtins.input", side_effect=[*deepcopy(self._args)]) as _:
                return func()
        else:
            return func(*deepcopy(self._args), **deepcopy(self._kwargs))

    def check(
        self,
        golden: Callable[..., Output],
        under_test: Callable[..., Output],
        metadata: TestMetadata,
    ) -> None:
        """Assert that `golden` and `under_test` give the same output.

        This method relies on `TestCase.assertEqual`, which will raise a unittest error
        if they differ. This means it can be plugged into any unittest TestCase as a
        helper method.
        """
        if metadata.check_stdout:
            golden_stdout, golden_output = self._eval(with_captured_stdout(golden))
            under_test_stdout, under_test_output = self._eval(
                with_captured_stdout(under_test)
            )

            self._assert_eq(
                golden_stdout,
                under_test_stdout,
                metadata,
                metadata.config.stdout_differ_msg,
            )

        else:
            golden_output = self._eval(golden)
            under_test_output = self._eval(under_test)

        self._assert_eq(
            golden_output, under_test_output, metadata, metadata.config.failure_msg
        )

    def _assert_eq(
        self, expected: Output, got: Output, metadata: TestMetadata, msg_format: str
    ) -> None:
        """Assert that expected equals got, formatting `msg_format` if not."""
        # we can only diff strings
        if isinstance(expected, str) and isinstance(got, str):
            diff_explanation = metadata.config.diff_explanation_msg
            diff = text_diff(got, expected)
        else:
            diff_explanation = ""
            diff = ""

        self.assertEqual(
            got,
            expected,
            msg=msg_format.format(
                input=repr(self),
                expected=repr(expected),
                output=repr(got),
                diff_explanation=diff_explanation,
                diff=diff,
            ),
        )

    def generate_test_case(
        self,
        golden: Callable[..., Output],
        under_test: Callable[..., Output],
        score: float,
        config: AgaConfig,
    ) -> AgaTestCase:
        """Generate a TestCase which tests `golden` against `under_test`."""
        name_fmt = config.test.name_fmt
        name_sep = config.test.name_sep

        name = self._name or name_fmt.format(
            args=self._args_repr(name_sep),
            kwargs=self._kwargs_repr(name_sep),
            sep=self._sep(name_sep),
        )
        metadata = TestMetadata(
            name=name,
            hidden=self._hidden,
            config=config.test,
            max_score=score,
            check_stdout=config.problem.check_stdout,
            mock_input=config.problem.mock_input,
        )
        return AgaTestCase(self, golden, under_test, metadata)

    def _args_repr(self, sep: str) -> str:
        return sep.join(repr(x) for x in self._args)

    def _kwargs_repr(self, sep: str) -> str:
        # we use k instead of repr(k) so we don't get quotes around it
        return sep.join(k + "=" + repr(v) for k, v in self._kwargs.items())

    def _sep(self, sep: str) -> str:
        """Return sep if both exist, "" otherwise."""
        assert sep == ","
        return self._args and self._kwargs and sep or ""

    def __repr__(self) -> str:
        args_repr = self._args_repr(",")
        kwargs_repr = self._kwargs_repr(",")
        sep = self._sep(",")

        return args_repr + sep + kwargs_repr

    def check_one(self, golden: Callable[..., Output]) -> None:
        """Check that the golden solution is correct."""
        if self._expect is not None:
            self.assertEqual(self._eval(golden), self._expect)


class _TestInputGroup(Generic[Output]):
    """A group of test cases with shared configuration."""

    def __init__(self, weight: int = 1, value: float = 0.0) -> None:
        self._test_cases: list[_TestInputs[Output]] = []
        self._prizes: list[Prize] = []
        self.score_info = ScoreInfo(weight, value)

    def add_test_case(self, case: _TestInputs[Output]) -> None:
        """Add a test case to the group."""
        self._test_cases.append(case)

    def add_prize(self, prize: Prize) -> None:
        """Add a prize to the group."""
        self._prizes.append(prize)

    def generate_test_suite(
        self,
        golden: Callable[..., Output],
        under_test: Callable[..., Output],
        group_score: float,
        config: AgaConfig,
    ) -> tuple[AgaTestSuite, list[ScoredPrize]]:
        """Generate a test suite from all the test cases for this group."""
        suite = AgaTestSuite(config, [])

        score_infos = [case.score_info for case in self._test_cases] + [
            prize.score_info for prize in self._prizes
        ]
        scores = compute_scores(score_infos, group_score)

        for (score, case) in zip(scores, self._test_cases):
            suite.addTest(case.generate_test_case(golden, under_test, score, config))

        scored_prizes = []
        for (score, prize) in zip(reversed(scores), reversed(self._prizes)):
            # reverse scores so we get the ones that correspond to the prizes
            scored_prizes.append(
                ScoredPrize(prize=prize, max_score=score)  # type: ignore
            )

        return suite, scored_prizes

    def check_one(self, golden: Callable[..., Output]) -> None:
        """Check the golden solution against all test cases."""
        for case in self._test_cases:
            case.check_one(golden)


class Problem(Generic[Output]):
    """Stores tests for a single problem."""

    def __init__(
        self,
        golden: Callable[..., Output],
        name: str,
        config: AgaConfig,
        is_script: bool,
    ) -> None:
        self._golden: Callable[..., Output] = golden
        self._name = name
        self._config = config
        self._ungrouped_prizes: list[Prize] = []
        self._ungrouped_tests: list[_TestInputs[Output]] = []
        self._groups: list[_TestInputGroup[Output]] = []
        self.is_script = is_script

    def add_test_case(
        self,
        *args: Any,
        aga_expect: Optional[Output] = None,
        aga_hidden: bool = False,
        aga_name: Optional[str] = None,
        aga_weight: int = 1,
        aga_value: float = 0.0,
        **kwargs: Any,
    ) -> None:
        """Add a test case to the current group.

        Student solutions will be checked against the golden solution; i.e., this method
        does _not_ produce a test of the golden solution.
        """
        case: _TestInputs[Output] = _TestInputs(
            *args,
            aga_expect=aga_expect,
            aga_hidden=aga_hidden,
            aga_name=aga_name,
            aga_weight=aga_weight,
            aga_value=aga_value,
            aga_mock_input=self._config.problem.mock_input,
            **kwargs,
        )
        self._ungrouped_tests.append(case)

    def add_prize(self, prize: Prize) -> None:
        """Add a prize to the current group."""
        self._ungrouped_prizes.append(prize)

    def add_group(self, grp: _TestInputGroup[Output]) -> None:
        """Add a group to the problem."""
        for case in self._ungrouped_tests:
            grp.add_test_case(case)

        for prize in self._ungrouped_prizes:
            grp.add_prize(prize)

        self._groups.append(grp)
        self._ungrouped_tests = []
        self._ungrouped_prizes = []

    def config(self) -> AgaConfig:
        """Get access to the problem's config."""
        return self._config

    def check(self) -> None:
        """Check that the problem is correct.

        Currently, this runs all tests of the golden solution.
        """
        for grp in self._virtual_groups():
            grp.check_one(self._golden)

    def generate_test_suite(
        self, under_test: Callable[..., Output], metadata: "SubmissionMetadata"
    ) -> tuple[AgaTestSuite, list[ScoredPrize]]:
        """Generate a `TestSuite` for the student submitted function.

        Neither the generated test suite nor the body of this function will run golden
        tests; instead, golden test cases are treated as equivalent to ordinary ones. To
        test the golden function, `check` should be used instead.

        Parameters
        ----------
        under_test : Callable[..., Output]
            The student submitted function.
        metadata : SubmissionMetadata
            The submission metadata.

        Returns
        -------
        AgaTestSuite
            A unittest test suite containing one test for each TestInput in this
            problem, checking the result of the problem's golden solution against
            `under_test`.
        list[ScorePrize]
            The prizes for the problem, with scores assigned.
        """
        ret_suite = AgaTestSuite(self._config, [])

        groups = self._virtual_groups()

        score_infos = [grp.score_info for grp in groups]
        scores = compute_scores(score_infos, metadata.total_score)

        ret_prizes = []
        for (score, grp) in zip(scores, groups):
            suite, prizes = grp.generate_test_suite(
                self._golden, under_test, score, self._config
            )
            ret_suite.addTest(suite)
            ret_prizes += prizes

        return ret_suite, ret_prizes

    def name(self) -> str:
        """Get the problem's name."""
        return self._name

    def expected_symbol(self) -> str:
        """Get the name of the symbol that should be tested against."""
        return self._golden.__name__

    def update_config_weak(self, config: AgaConfig) -> None:
        """Update any non-default items in self.config."""
        self._config.update_weak(config)

    def _virtual_groups(self) -> list[_TestInputGroup[Output]]:
        """Get the list of groups, plus the current group under construction.

        We need to do it this way because while the problem is being read we don't know
        the configuration of the last test group yet.
        """
        if self._ungrouped_tests or self._ungrouped_prizes:
            virtual_group: _TestInputGroup[Output] = _TestInputGroup()

            for case in self._ungrouped_tests:
                virtual_group.add_test_case(case)

            for prize in self._ungrouped_prizes:
                virtual_group.add_prize(prize)

            return self._groups + [virtual_group]

        else:
            return self._groups


def problem(
    name: Optional[str] = None,
    script: bool = False,
    check_stdout: Optional[bool] = None,
    mock_input: Optional[bool] = None,
) -> Callable[[Callable[..., Output]], Problem[Output]]:
    """Declare a function as the golden solution to a problem.

    This method should decorate a function which is known to produce the correct
    outputs, which we refer to as the "golden solution". It also provides facilities
    for testing that solution, via golden test cases, constructed by passing the
    output argument to the test_case decorator.

    Parameters
    ----------
    name : Optional[str]
        The problem's name. If None (the default), the wrapped function's name will be
        used.
    script : bool
        Whether the problem represents a script, as opposed to a function. Implies
        `check_stdout` and `mock_input` unless they are passed explicitly.
    check_stdout : Optional[bool]
        Overrides the `problem.check_stdout` configuration option. If True, check the
        golden solution's stdout against the student submission's for all test cases.
    mock_input : Optional[bool]
        Overrides the `problem.mock_input` configuration option. If True, test cases for
        this problem will be interpreted as mocked outputs of `builtins.input`, rather
        than inputs to the function.

    Returns
    -------
    Callable[[Callable[..., T]], Problem[T]]
        A decorator which turns a golden solution into a problem.
    """
    config = AgaConfig()

    if check_stdout is not None:
        config.problem.check_stdout = check_stdout
        config.problem.check_stdout_overridden = True

    if mock_input is not None:
        config.problem.mock_input = mock_input
        config.problem.mock_input_overridden = True

    def outer(func: Callable[..., Output]) -> Problem[Output]:
        problem_name = name or func.__name__

        if script:
            if check_stdout is None:
                config.problem.check_stdout = True
                config.problem.check_stdout_overridden = True

            if mock_input is None:
                config.problem.mock_input = True
                config.problem.mock_input_overridden = True

        return Problem(func, problem_name, config, script)

    return outer


def _check_reserved_keyword(kwd: str) -> None:
    """Raise an error if `kwd` is reserved."""
    if kwd.startswith("aga_") and kwd not in (
        "aga_expect",
        "aga_hidden",
        "aga_name",
        "aga_weight",
        "aga_value",
    ):
        raise ValueError(
            f'invalid keyword arg "{kwd}" to `test_case`: all keyword args '
            "beginning `aga_` are reserved"
        )


def test_case(
    *args: Any,
    **kwargs: Any,
) -> Callable[[Problem[Output]], Problem[Output]]:
    r"""Declare a specific test case for some problem.

    Parameters
    ----------
    args :
        The arguments to be passed to the functions under test.
    aga_expect : Optional[T]
        If aga_expect is None, the inputs will be tested against the wrapped function,
        the "golden solution" to the problem. If aga_expect is specified, the inputs
        will double as a test _of_ the golden solution; to successfully produce the
        problem grader, the golden solution must return aga_expect from the given input.
    aga_hidden : bool
        If True, hide the problem from students on supported frontends.
    aga_name : Optional[str]
        The test case's name. If `None`, defaults to "Test {inputs}", where {inputs} is
        a comma-separated list of args and kwargs.
    aga_weight : int
        The test case's relative weight to the group's score. See :ref:`Determining
        Score` for details.
    aga_value : int
        The test case's absolute score. See :ref:`Determining Score` for details.
    kwargs :
        Keyword arguments to be passed to the functions under test. Any keyword starting
        with aga\_ is reserved.

    Returns
    -------
    Callable[[Problem[T]], Problem[T]]
        A decorator which adds the test case to a problem.
    """
    for kwd, _ in kwargs.items():
        _check_reserved_keyword(kwd)

    def outer(prob: Problem[Output]) -> Problem[Output]:
        prob.add_test_case(
            *args,
            **kwargs,
        )
        return prob

    return outer


def test_cases(
    *args: Iterable[Any],
    aga_product: bool = True,
    **kwargs: Iterable[Any],
) -> Callable[[Problem[Output]], Problem[Output]]:
    r"""Generate many test cases programatically, from generators of inputs.

    Parameters
    ----------
    args :
        Generators for arguments to be passed to the functions under test.
    aga_product : bool
        Whether to take a cartesian product of the generators (creating one test case
        for each set of inputs in the product), or to zip them (iterate through each
        generator in sequence). Default `True`.
    aga_*
        Other `aga_` keywords have their meaning inherited from `test_case`, and are
        applied to each test case generated by this function. Aga_expect is not
        supported.
    kwargs :
        Generators for keyword arguments to be passed to the functions under test. Any
        keyword starting with aga\_ is reserved.

    Returns
    -------
    Callable[[Problem[T]], Problem[T]]
        A decorator which adds the test cases to a problem.
    """

    def outer(prob: Problem[Output]) -> Problem[Output]:
        combinator = product if aga_product else zip

        combined_args = combinator(*args)
        combined_kwargs = combinator(*kwargs.values())

        for curr_args, curr_kwargs in combinator(combined_args, combined_kwargs):
            # we have to reassemble a kwargs dictionary that actually has keys
            final_kwargs = dict(zip(kwargs.keys(), curr_kwargs))

            # i think we have to enumerate all of the arguments to test_case by
            # hand; we can't capture them separately from the test case kwargs
            prob = test_case(
                *curr_args,
                **final_kwargs,
            )(prob)

        return prob

    return outer


def group(
    weight: int = 1, value: float = 0.0
) -> Callable[[Problem[Output]], Problem[Output]]:
    """Declare a group of problems.

    Parameters
    ----------
    weight : int
        The group's relative weight to the problem's score. See :ref:`Determining Score`
        for details.
    value : int
        The group's absolute score. See :ref:`Determining Score` for details.

    Returns
    -------
    Callable[[Problem[T]], Problem[T]]
        A decorator which adds the group to a problem.
    """

    def outer(prob: Problem[Output]) -> Problem[Output]:
        prob.add_group(_TestInputGroup(weight, value))
        return prob

    return outer



"""Add points prizes to problems.

This module contains the `prize` decorator, which lets you define custom post-test-run
points hooks for things like correctness and lateness. It also contains several prizes,
defined for convenience.
"""
from .core import SubmissionMetadata
from .runner import TcOutput
from .score import PrizeCriteria, all_correct, correct_and_on_time, on_time, prize

__all__ = (
    "prize",
    "all_correct",
    "on_time",
    "correct_and_on_time",
    "PrizeCriteria",
    "SubmissionMetadata",
    "TcOutput",
)



"""Utilities for aga."""
import difflib
from traceback import extract_tb


def text_diff(old: str, new: str) -> str:
    """Generate a diff between old and new."""
    old_list = old.splitlines(keepends=True)
    new_list = new.splitlines(keepends=True)

    return "".join(difflib.ndiff(old_list, new_list))


def limited_traceback(traceback) -> str:  # type: ignore
    """Format a traceback, including removing aga-specific parts of the trace."""
    # this is essentially a list of FrameSummary objects
    stack_summary = extract_tb(traceback)

    # this is a hack, but the idea is that if the file path has `aga` or `unittest` in
    # it, it's probably part of our infrastructure, rather than the student submission.
    out = ""
    for (frame, formatted_frame) in zip(stack_summary, stack_summary.format()):
        if "aga" in frame.filename or "unittest" in frame.filename:
            # reset output, we don't want any earlier lines bc student code hasn't been
            # called yet
            out = ""
        else:
            out += "\n"
            out += formatted_frame

    return out



"""A frontend layer for producing zip files for the gradescope autograder.

The main work of this module is to generate the gradescope autograder zip file. The
specification for this file is here:
<https://gradescope-autograders.readthedocs.io/en/latest/specs/>. Specifically, we are
given two entrypoints:

 * setup.sh, a Bash script.
 * run_autograder, any executable file.

Both will be executed from a directory `/autograder` created by the autograder's Docker
harness.

We may also add arbitrary files to the zip. These will be available at
`/autograder/source`.

Docker will provide the student submission in the `/autograder/submission` directory.

Our current high-level approach is as follows:

 * Install python and aga in the `setup.sh` file.
 * Pickle the `Problem` at `/autograder/source/problem.pckl`.
 * Provide a `run_autograder` python file which unpickles the `Problem`, uses the
 `loader` module to load the student submission, and runs them against each other.

Some implementation details:

We use tempfile from the Python standard library to generate a temporary directory in
which we build the contents of the zip archive. The `run_autograder` and `setup.sh`
files in `resources/` are copied directly to this tempdir.

We use `dill` for pickling, instead of the standard library's `pickle`, because pickle
behaves weirdly with decorated functions, and `functools.update_wrapper` was
insufficient to fix this.
"""

from .into_zip import InvalidProblem, into_gradescope_zip

__all__ = ("into_gradescope_zip", "InvalidProblem")



"""The command-line UI."""

from rich import print as rprint
from rich.box import ROUNDED
from rich.panel import Panel

from ..runner import ProblemOutput

RICH_PANEL_OPTS = {
    "box": ROUNDED,
    "title_align": "left",
    "subtitle_align": "right",
}


def print_fancy_summary(output: ProblemOutput) -> None:
    """Print a fancy summary of the problem."""
    rprint(
        Panel(
            output.output,
            title="Overall",
            subtitle=f"Total score: {output.score}",
            **RICH_PANEL_OPTS,  # type: ignore
        )
    )

    for test in output.tests:
        if not test.hidden:
            rprint(
                Panel(
                    test.output or "",
                    title=("" if test.is_correct() else "[bright_red]") + test.name,
                    subtitle=f"{test.score}/{test.max_score}",
                    **RICH_PANEL_OPTS,  # type: ignore
                )
            )



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['aga', 'aga.cli', 'aga.gradescope', 'aga.gradescope.resources']

package_data = \
{'': ['*']}

install_requires = \
['dacite>=1.6.0,<2.0.0',
 'dataclasses-json>=0.5.6,<0.6.0',
 'dill>=0.3.4,<0.4.0',
 'gradescope-utils>=0.4.0,<0.5.0',
 'toml>=0.10.2,<0.11.0',
 'typer[all]>=0.6.1,<0.7.0',
 'types-toml>=0.10.8,<0.11.0']

entry_points = \
{'console_scripts': ['aga = aga.cli:aga_app']}

setup_kwargs = {
    'name': 'aga',
    'version': '0.8.6',
    'description': 'aga grades assignments',
    'long_description': '<div align="center">\n\n# aga grades assignments\n\n[![tests](https://github.com/nihilistkitten/aga/workflows/tests/badge.svg)](https://github.com/nihilistkitten/aga/actions?workflow=tests)\n[![lints](https://github.com/nihilistkitten/aga/workflows/lints/badge.svg)](https://github.com/nihilistkitten/aga/actions?workflow=lints)\n[![Codecov](https://codecov.io/gh/nihilistkitten/aga/branch/main/graph/badge.svg)](https://codecov.io/gh/nihilistkitten/aga)\n[![PyPI](https://img.shields.io/pypi/v/aga.svg)](https://pypi.org/project/aga/)\n[![Read the Docs](https://readthedocs.org/projects/aga/badge/)](https://aga.readthedocs.io/)\n[![License](https://img.shields.io/github/license/nihilistkitten/aga)](https://choosealicense.com/licenses/mit/)\n\n</div>\n\n**aga** (**a**ga **g**rades **a**ssignments) is a tool for easily producing autograders for python programming assignments, originally developed for Reed College\'s CS1 course.\n\n## Motivation\n\nUnlike traditional software testing, where there is likely no _a priori_ known-correct implementation, there is always such an implementation (or one can be easily written by course staff) in homework grading. Therefore, applying traditional software testing frameworks to homework grading is limited. Relying on reference implementations (what aga calls _golden solutions_) has several benefits:\n\n1. Reliability: having a reference solution gives a second layer of confirmation for the correctness of expected outputs. Aga supports _golden tests_, which function as traditional unit tests of the golden solution.\n2. Test case generation: many complex test cases can easily be generated via the reference solution, instead of needing to work out the expected output by hand. Aga supports generating test cases from inputs without explcitly referring to an expected output, and supports collecting test cases from python generators.\n3. Property testing: unit testing libraries like [hypothesis](https://hypothesis.readthedocs.io) allow testing large sets of arbitrary inputs for certain properties, and identifying simple inputs which reproduce violations of those properties. This is traditionally unreliable, because identifying specific properties to test is difficult. In homework grading, the property can simply be "the input matches the golden solution\'s output." Support for hypothesis is a [long-term goal](https://github.com/nihilistkitten/aga/issues/32) of aga.\n\n## Installation\n\nInstall from pip:\n\n```bash\npip install aga\n```\n\nor with the python dependency manager of your choice (I like [poetry](https://github.com/python-poetry/poetry)).\n\n## Example\n\nIn `square.py` (or any python file), write:\n\n```python\nfrom aga import problem, test_case, test_cases\n\n@test_cases([-3, 100])\n@test_case(2, aga_expect=4)\n@test_case(-2, aga_expect=4)\n@problem()\ndef square(x: int) -> int:\n    """Square x."""\n    return x * x\n```\n\nThen run `aga gen square.py` from the directory with `square.py`. This will generate a ZIP file suitable for upload to Gradescope.\n\n## Usage\n\nAga relies on the notion of a _golden solution_ to a given problem which is known to be correct. The main work of the library is to compare the output of this golden solution on some family of test inputs against the output of a student submission. To that end, aga integrates with frontends: existing classroom software which allow submission of student code. Currently, only Gradescope is supported.\n\nTo use aga:\n\n1. Write a golden solution to some programming problem.\n2. Decorate this solution with the `problem` decorator.\n3. Decorate this problem with any number of `test_case` decorators, which take arbitrary positional or keyword arguments and pass them verbatim to the golden and submitted functions.\n4. Generate the autograder using the CLI: `aga gen <file_name>`.\n\nThe `test_case` decorator may optionally take a special keyword argument called `aga_expect`. This allows easy testing of the golden solution: aga will not successfully produce an autograder unless the golden solution\'s output matches the `aga_expect`. You should use these as sanity checks to ensure your golden solution is implemented correctly.\n\nFor more info, see the [tutorial](https://aga.readthedocs.io/en/stable/tutorial.html).\n\nFor complete documentation, including configuring problem and test case metadata, see the [API reference](https://aga.readthedocs.io/en/stable/reference.html).\n\nFor CLI documentation, run `aga --help`, or access the docs [online](https://aga.readthedocs.io/en/stable/cli.html).\n\n## Contributing\n\nBug reports, feature requests, and pull requests are all welcome. For details on our test suite, development environment, and more, see the [developer documentation](https://aga.readthedocs.io/en/stable/development.html).\n\n<!-- vim:set tw=0: -->\n',
    'author': 'Riley Shahar',
    'author_email': 'riley.shahar@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/nihilistkitten/aga',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.10.0,<4.0.0',
}


setup(**setup_kwargs)



"""Utilities for easily computing problem score."""

from dataclasses import dataclass
from typing import TYPE_CHECKING, Callable

if TYPE_CHECKING:
    from .core import Output, Problem, SubmissionMetadata
    from .runner import TcOutput


PrizeCriteria = Callable[[list["TcOutput"], "SubmissionMetadata"], tuple[float, str]]


@dataclass(frozen=True)
class ScoreInfo:
    """Info to help compute a score.

    Attributes
    ----------
    value : float
        The object's absolute point value.
    weight : int
        The relative of the remaining score this object will claim.
    """

    weight: int
    value: float


def compute_scores(score_infos: list[ScoreInfo], total_score: float) -> list[float]:
    """Compute the scores of a list of scorable objects.

    Returns their scores in the same order as the initial list.
    """
    out = []
    total_weight = 0
    for s in score_infos:
        total_score -= s.value
        total_weight += s.weight

    for s in score_infos:
        if total_weight > 0:
            out.append(s.value + (total_score * s.weight) / total_weight)
        else:
            out.append(s.value)

    return out


@dataclass(frozen=True)
class Prize:
    """A points prize, which can be earned by meeting certain problem-wide criteria."""

    name: str
    criteria: PrizeCriteria
    score_info: ScoreInfo


@dataclass(frozen=True)
class ScoredPrize:
    """A prize with attached score."""

    prize: Prize
    max_score: float


def prize(
    criteria: PrizeCriteria,
    name: str = "Prize",
    weight: int = 1,
    value: float = 0.0,
) -> Callable[["Problem[Output]"], "Problem[Output]"]:
    """Add a points prize to the problem.

    Parameters
    ----------
    criteria : Callable[[list[TcOutput], SubmissionMetadata], tuple[float, str]
        The criteria for awarding the prize's points. The first returned value should be
        a float from 0 to 1 which determines the fraction of points to assign. The
        second should be a string which will be displayed to the student.
    name : str
        The name of the prize, to be displayed to the student.
    weight : int
        The prize's weight. See :ref:`Determining Score` for details.
    value : int
        The prize's absolute score. See :ref:`Determining Score` for details.

    Returns
    -------
    Callable[[Problem[T]], Problem[T]]
        A decorator which adds the prize to a problem.
    """
    to_add = Prize(name, criteria, ScoreInfo(weight, value))

    def inner(problem: "Problem[Output]") -> "Problem[Output]":
        problem.add_prize(to_add)
        return problem

    return inner


def all_correct(tests: list["TcOutput"], _: "SubmissionMetadata") -> tuple[float, str]:
    """1.0 if all tests passed, 0.0 otherwise.

    For use as a prize.
    """
    if all(t.is_correct() for t in tests):
        return 1.0, "Good work! You earned these points since all tests passed."
    else:
        return 0.0, "To earn these points, make sure all tests pass."


def on_time(_: list["TcOutput"], metadata: "SubmissionMetadata") -> tuple[float, str]:
    """1.0 if the submission was on time, 0.0 otherwise.

    For use as a prize.
    """
    if metadata.is_on_time():
        return (
            1.0,
            "Good work! You earned these points by turning the assignment in on time.",
        )
    else:
        return (
            0.0,
            "To earn this next time, make sure to turn the assignment in on time!",
        )


def correct_and_on_time(
    tests: list["TcOutput"], metadata: "SubmissionMetadata"
) -> tuple[float, str]:
    """1.0 if the submission was correct and passed all tests, 0.0 otherwise.

    For use as a prize.
    """
    match (bool(all_correct(tests, metadata)[0]), bool(on_time(tests, metadata)[0])):
        case (True, True):
            return (
                1.0,
                "Good work! You earned these points since all tests passed and "
                "you turned in the assignment on time.",
            )
        case (False, True):
            return 0.0, "To earn these points, make sure all tests pass."
        case (True, False):
            return (
                0.0,
                "To earn these points next time, "
                "make sure to turn the assignment in on time.",
            )
        case _:
            return (
                0.0,
                "To earn these points next time, "
                "make sure to turn the assignment in on time, and that all tests pass.",
            )



"""The main command-line typer application."""

from datetime import datetime
from typing import Iterable, Optional, Tuple

import typer

from ..config import AgaConfig, load_config_from_path
from ..core import Output, Problem, SubmissionMetadata
from ..gradescope import into_gradescope_zip
from ..loader import load_problems_from_path
from ..runner import load_and_run
from .ui import print_fancy_summary

app = typer.Typer()

FRONTENDS = (
    (
        "gradescope",
        "generate a gradescope autograder zip",
    ),
)


def complete_frontend(incomplete: str) -> Iterable[Tuple[str, str]]:
    """Autocomplete a frontend."""
    for frontend in FRONTENDS:
        if frontend[0].startswith(incomplete):
            yield frontend


def _gen_gradescope(problem: Problem[Output], path: Optional[str] = None) -> None:
    """Generate a Gradescope autograder zip."""
    zip_path = into_gradescope_zip(problem, path=path)
    typer.echo(zip_path)


def _load_config(path: str = "aga.toml") -> AgaConfig:
    """Load a config file from path, or return the default."""
    try:
        config = load_config_from_path(path)
    except FileNotFoundError:
        config = AgaConfig()

    return config


def _load_problem(path: str, config: AgaConfig) -> Problem[Output]:
    """Load a problem from the top-level directory."""
    problems = list(load_problems_from_path(path))

    if not problems:
        typer.echo(f"No problems found at {path}.", err=True)
        raise typer.Exit(1)

    if len(problems) > 1:
        typer.echo(
            f"Multiple problems found in {path}. "
            "Currently, only one problem is supported per file.",
            err=True,
        )
        raise typer.Exit(1)

    problem = problems[0]
    problem.update_config_weak(config)

    return problem


def _handle_invalid_frontend(frontend: str) -> None:
    """Error on an invalid frontend."""
    typer.echo(f"invalid frontend: {frontend}", err=True)
    raise typer.Exit(1)


def _check_problem(problem: Problem[Output]) -> None:
    """Check that problem is valid."""
    try:
        problem.check()
    except AssertionError as err:
        typer.echo(f"{problem.name()} failed some golden tests: {err}", err=True)
        raise typer.Exit(1) from err


@app.command()
def gen(
    source: str = typer.Argument(
        ...,
        help="The file with the problem to generate.",
    ),
    frontend: str = typer.Option(
        "gradescope",
        "-f",
        "--frontend",
        autocompletion=complete_frontend,
        help="The frontend to use. Currently only gradescope is supported.",
    ),
    output: Optional[str] = typer.Option(
        None,
        "--output",
        "-o",
        help="The path to place the output file(s).",
    ),
    config_file: str = typer.Option(
        "aga.toml", "--config", "-c", help="The path to the aga config file."
    ),
) -> None:
    """Generate an autograder file for a problem."""
    config = _load_config(config_file)
    problem = _load_problem(source, config)  # type: ignore
    _check_problem(problem)

    if frontend == "gradescope":
        _gen_gradescope(problem, output)
    else:
        _handle_invalid_frontend(frontend)


@app.command()
def check(
    source: str = typer.Argument(
        ...,
        help="The problem to check.",
    ),
    config_file: str = typer.Option(
        "aga.toml", "--config", "-c", help="The path to the aga config file."
    ),
) -> None:
    """Check a problem against test cases with an `aga_expect`."""
    config = _load_config(config_file)
    problem = _load_problem(source, config)  # type: ignore
    _check_problem(problem)
    typer.echo(f"{problem.name()} passed golden tests.")


@app.command()
# pylint: disable=too-many-arguments
def run(
    source: str = typer.Argument(
        ...,
        help="The problem to run.",
    ),
    submission: str = typer.Argument(
        ..., help="The file containing the student submission."
    ),
    config_file: str = typer.Option(
        "aga.toml", "--config", "-c", help="The path to the aga config file."
    ),
    points: float = typer.Option(
        20.0, "--points", help="The total number of points for the problem."
    ),
    due: datetime = typer.Option(
        datetime.now,
        "--due",
        help="The problem due date.",
        show_default="now",  # type: ignore
    ),
    submitted: datetime = typer.Option(
        datetime.now,
        "--submitted",
        help="The problem submission date.",
        show_default="now",  # type: ignore
    ),
    previous_submissions: int = typer.Option(
        0, "--previous_submissions", help="The number of previous submissions."
    ),
) -> None:
    """Run the autograder on an example submission."""
    config = _load_config(config_file)
    problem: Problem = _load_problem(source, config)  # type: ignore
    metadata = SubmissionMetadata(points, submitted - due, previous_submissions)
    result = load_and_run(problem, submission, metadata)
    print_fancy_summary(result)


click_object = typer.main.get_command(app)  # exposed for documentation
